import typing

import abc

from smallab.smallab_types import Specification


class ExperimentBase(abc.ABC):
    def set_logger_name(self, logger):
        """
        Called by ExperimentRunner to set a per-experiment logger
        :param logger: The logger which all logging in this experiment should be done to
        """
        self.logger = logger

    def get_logger_name(self):
        """
        Gets the logger that all logging in this experiment should be done to
        :return: An instance of logging.logger which corresponds to this experiment
        """
        return self.logger

    def set_logging_folder(self, folder_loc):
        """
        Set by ExperimentLogger, where the log files should go.
        :param folder_loc:
        :return:
        """
        self.logging_folder = folder_loc

    def get_logging_folder(self):
        """
        Read by ExperimentLogger, where the log files should go.
        :return:
        """
        return self.logging_folder


class Experiment(ExperimentBase):
    """
    This represents the base class for what the user needs to implement.
    The main method takes a dictionary which is the specification and returns a dictionary which is the result.
    This is run by the ExperimentRunner class

    The lifecycle of this object is
    {} is called internally by smallab
    {.set_logging_folder() -> .get_logging_folder() -> .set_logger()} -> .main()
    """

    @abc.abstractmethod
    def main(self, specification: Specification) -> typing.Dict:
        """
        The method that should be overriden when using smallab.
        Should contain the code for running your experiment.

        If an exception is raised the experiment is considered failed.

        :param specification: A dictionary passed which represents the variables of this experiment
        :return: A dictionary of results which will be serialized and saved if this experiment is succesful
        """
        pass



import json
import pickle

import os

from smallab.file_locations import get_experiment_save_directory


def experiment_iterator(name):
    for root, _, files in os.walk(get_experiment_save_directory(name)):
        for fname in files:
            if ".pkl" in fname:
                with open(os.path.join(root, fname), "rb") as f:
                    yield pickle.load(f)
            if ".json" in fname and fname != "specification.json":
                with open(os.path.join(root, fname), "r") as f:
                    yield json.load(f)



import typing

import abc

from smallab.experiment_types.experiment import ExperimentBase
from smallab.smallab_types import Specification, ExpProgressTuple


class OverlappingOutputCheckpointedExperimentReturnValue():
    def __init__(self, should_continue: bool, specification: Specification, return_value: typing.Dict,
                 progress: typing.SupportsFloat, max_iterations: typing.SupportsFloat):
        """
        The holder class for output from an OverlappingOutputCheckpointedExperiment
        :param should_continue: True if the experiment needs to continue (to perform more iterations etc)
        :param specification: The specification this experiment will be saved under (not necessarily the specification passed to the experiment)
        :param return_value: The return value from this output
        :param progress: The progress (out of max_iterations)
        :param max_iterations: The maximum amount of iterations that will be run
        """
        self.should_continue = should_continue
        self.specification = specification
        self.return_value = return_value
        self.progress = progress
        self.max_iterations = max_iterations


class OverlappingOutputCheckpointedExperiment(ExperimentBase):
    """
    This experiment class handles the case where you have an experiment that you would like multiple overlapping experiments saved
    For example, you want to run your method for 100,200 and 300 iterations.
    This class allows you to save the 100 and 200 iteration experiments without restarting from scratch (for the 200 iteration experiment
    it would continue from the 100 iteration experiment)

    The experiment is also checkpointed between iterations you want to save like CheckpointedExperiment (and thus must
    also be picklable).
    """

    @abc.abstractmethod
    def initialize(self, specification: Specification):
        pass

    @abc.abstractmethod
    def step(self) -> typing.Union[ExpProgressTuple, OverlappingOutputCheckpointedExperimentReturnValue]:
        pass



import json
import logging

import dill
import os
import types
from copy import deepcopy

from smallab.dashboard.dashboard_events import BeginEvent, CompleteEvent, FailedEvent
from smallab.dashboard.utils import put_in_event_queue
from smallab.experiment_types.handlers.registry import run_with_correct_handler
from smallab.file_locations import (get_json_file_location, get_save_file_directory, get_pkl_file_location,
                                    get_specification_file_location, get_log_file)
from smallab.specification_hashing import specification_hash


def save_run(name, experiment, specification, result, force_pickle):
    os.makedirs(get_save_file_directory(name, specification))
    output_dictionary = {"specification": specification, "result": result}
    json_serialize_was_successful = False
    # Try json serialization
    if not force_pickle:
        json_filename = get_json_file_location(name, specification)
        try:
            with open(json_filename, "w") as f:
                json.dump(output_dictionary, f)
            json_serialize_was_successful = True
        except Exception:
            logging.getLogger(experiment.get_logger_name()).warning("Json serialization failed with exception",
                                                                    exc_info=True)
            os.remove(json_filename)
    # Try pickle serialization
    if force_pickle or not json_serialize_was_successful:
        pickle_file_location = get_pkl_file_location(name, specification)
        specification_file_location = get_specification_file_location(name, specification)
        try:
            with open(pickle_file_location, "wb") as f:
                dill.dump(output_dictionary, f)
            with open(specification_file_location, "w") as f:
                json.dump(specification, f)
        except Exception:
            logging.getLogger(experiment.get_logger_name()).critical("Experiment results serialization failed!!!",
                                                                     exc_info=True)
            try:
                os.remove(pickle_file_location)
            except FileNotFoundError:
                pass
            try:
                os.remove(specification_file_location)
            except FileNotFoundError:
                pass


def run_and_save(name, experiment, specification, propagate_exceptions, callbacks, force_pickle):
    experiment = deepcopy(experiment)
    specification_id = specification_hash(specification)
    logger_name = "smallab.{specification_id}".format(specification_id=specification_id)
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler(get_log_file(experiment, specification_id))
    formatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    experiment.set_logger_name(logger_name)
    put_in_event_queue(BeginEvent(specification_id))

    def _interior_fn():
        result = run_with_correct_handler(experiment, name, specification)
        if isinstance(result, types.GeneratorType):
            for cur_result in result:
                save_run(name, experiment, cur_result["specification"], cur_result["result"], force_pickle)
        else:
            save_run(name, experiment, specification, result, force_pickle)
        for callback in callbacks:
            callback.on_specification_complete(specification, result)
        return None

    if not propagate_exceptions:
        try:
            _interior_fn()

            put_in_event_queue(CompleteEvent(specification_id))
        except Exception as e:
            logger.error("Specification Failure", exc_info=True)
            put_in_event_queue(FailedEvent(specification_id))
            for callback in callbacks:
                callback.on_specification_failure(e, specification)
            return e
    else:
        _interior_fn()
        put_in_event_queue(CompleteEvent(specification_id))
        return None



import typing

import abc

from smallab.experiment_types.experiment import ExperimentBase
from smallab.smallab_types import Specification, ExpProgressTuple


class CheckpointedExperiment(ExperimentBase):
    """
    CheckpointedExperiment is an Experiment which can be stopped and restarted.
    Step is called multiple times, after it returns the current experiment is serialized allowing the experiment to
    restart seamlessly since to step it will not look any different whether python has shut down in between successive calls.
    Smallab will handle serialization and deserialization of this object.
    This will fail if objects attached to the experiment are not pickleable.

    The lifecycle of this object is

    {} is called internally by smallab, []* is called multiple times
    {.set_logging_folder() -> .get_logging_folder() -> .set_logger()} -> .initialize() -> [.step()]*
    """

    @abc.abstractmethod
    def initialize(self, specification: Specification):
        pass

    @abc.abstractmethod
    def step(self) -> typing.Union[ExpProgressTuple, typing.Dict]:
        """
        This method is to do work between checkpoints.
        Intermediate results should be saved to self
        :return: A tuple to indicate how much progress has been made if the experimetn is not done or  a dictionary of results when experiment has finished
        """
        pass



import typing

from smallab.experiment_types.checkpointed_experiment import CheckpointedExperiment
from smallab.experiment_types.experiment import ExperimentBase, Experiment
from smallab.experiment_types.handlers.checkpointed_experiment_handler import CheckpointedExperimentHandler
from smallab.experiment_types.handlers.overlapping_output_checkpointed_experiment_handler import \
    OverlappingOutputCheckpointedExperimentHandler
from smallab.experiment_types.overlapping_output_experiment import OverlappingOutputCheckpointedExperiment
from smallab.smallab_types import Specification


def run_with_correct_handler(experiment: ExperimentBase, name: typing.AnyStr, specification: Specification):
    if isinstance(experiment, CheckpointedExperiment):
        return CheckpointedExperimentHandler().run(experiment, name, specification)
    elif isinstance(experiment, OverlappingOutputCheckpointedExperiment):
        return OverlappingOutputCheckpointedExperimentHandler().run(experiment, name, specification)
    elif isinstance(experiment, Experiment):
        return experiment.main(specification)
    else:
        raise Exception("Experiment Handler not Understood")



#Subclassing experiment is the simplest way to implement your experiment but is not fault tolerant
#Subclassing CheckpointedExperiment allows your experiment to fail (or the running machine to lose power) and resume at a later point
#Note: Checkpointing does have some overhead so if the experiment is very fast consider not using this
import logging
import time
import typing

import random
from numpy.random.mtrand import RandomState

from smallab.experiment_types.checkpointed_experiment import CheckpointedExperiment
from smallab.experiment_types.experiment import Experiment
from smallab.runner.runner import ExperimentRunner
from smallab.runner_implementations.multiprocessing_runner import MultiprocessingRunner
from smallab.smallab_types import Specification
from smallab.specification_generator import SpecificationGenerator
from examples.example_utils import delete_experiments_folder


#Similar experiment as before but now it will sleep for a bit, taking longer to complete in serial
class SimpleExperiment(CheckpointedExperiment):
    #Initialize will be called once to set up your experiment
    #Each specification will have one instance of your experiment saved with it, so save variables you need to self
    def initialize(self, specification: Specification):
        self.seed = specification['seed']
        self.num_calls = specification["num_calls"]
        logging.getLogger(self.get_logger_name()).info("Doing work!")
        self.rs = RandomState(self.seed)
        self.i = 0

    #Step will be called many times until it does not return None
    def step(self):
        logging.getLogger(self.get_logger_name()).info("Stepping")
        self.i += 1
        if self.i < self.num_calls:
            self.r = self.rs.random()
            time.sleep(10 * self.r)
        #this experiment can have a random transient failure!
        #Since it's checkpointed, it will likely succeed after running it again
        if random.randint(0,100) > 90:
            raise Exception("Something bad happened, a moth flew into the computer!")
        if self.i >= self.num_calls:
            #Done with the experiment, return the results dictionary like normal
            return {"number": self.r}
        else:
            #This experiment isn't done, return the progress as a tuple to update the dashboard
            return (self.i, self.num_calls)


#Same specification as before
generation_specification = {"seed": [1, 2, 3, 4, 5, 6, 7, 8], "num_calls": [1, 2, 3]}
specifications = SpecificationGenerator().generate(generation_specification)

name = "checkpointed_run"
#This time we will run them all in parallel
runner = ExperimentRunner()
runner.run(name, specifications, SimpleExperiment(),specification_runner=MultiprocessingRunner())

#Some of our experiments may have failed, let's call run again to hopefully solve that
runner.run(name, specifications, SimpleExperiment(),specification_runner=MultiprocessingRunner())

#Cleanup example
delete_experiments_folder(name)


import typing

import abc

from smallab.experiment_types.experiment import ExperimentBase
from smallab.smallab_types import Specification


class BaseHandler(abc.ABC):

    @abc.abstractmethod
    def run(self, experiment: ExperimentBase, name: typing.AnyStr, specification: Specification):
        pass
